#version 420 core

const int MAX_POINT_LIGHTS = 3;                                                     
const int MAX_SPOT_LIGHTS = 4; 

in vec3 v_position;
in vec3 v_worldNormal;
in vec3 v_view;
in vec2 fragmentUV;
in vec4 v_smPosition;
in vec3 v_tangent; 

layout (location = 0) out vec4 FragColor;
layout (location = 1) out vec4 BrightColor;

	//Object Uniforms
uniform vec2 v_nearFar;
uniform sampler2D i_textureUnit;
uniform sampler2D i_normalUnit; 
uniform sampler2DShadow i_shadowMaptextureUnit;
uniform samplerCube i_shadowMapcubeUnit;
uniform vec3 v_matColor;
uniform float f_matAmbient;
uniform float f_matDiffuse;
uniform float f_matSpecular;
uniform float f_matShininess;
uniform float f_alpha;

	//Light Uniforms
struct AmbientLight                                                             
{                                                                                   
    vec3 v_color;                                                                                                                           
};  

struct DirectionalLight                                                             
{                                                                                   
    vec3 v_color;                                                              
    vec3 v_direction;                                                                 
};  

struct PointLight                                                             
{                                                                                   
    vec3 v_color;                                                              
    vec3 v_position;   
	vec3 v_attenuation;                                                               
}; 

struct SpotLight                                                             
{                                                                                   
    vec3 v_color;                                                              
    vec3 v_position;   
	vec3 v_attenuation; 
	vec3 v_direction;
	float f_cutoff;                                                                 
}; 

uniform int numPointLights;                                                                
uniform int numSpotLights;  
uniform AmbientLight ambLight;
uniform DirectionalLight dirLight;
uniform PointLight pointLights[MAX_POINT_LIGHTS];
uniform SpotLight spotLights[MAX_SPOT_LIGHTS];

vec3 CalcBumpedNormal();
vec3 DirectionalLightsIllumination(vec3 N, vec3 objColor);
vec3 PointLightsIllumination(vec3 V, vec3 N, vec3 objColor);
vec3 SpotLightsIllumination(vec3 V, vec3 N, vec3 objColor);
vec3 calcPointLightIntensity(int lightIndex);
vec3 calcSpotLightIntensity(int lightIndex);

vec3 cubeShadowSamplingDirection[20] = vec3[]
(
   vec3( 1,  1,  1), vec3( 1, -1,  1), vec3(-1, -1,  1), vec3(-1,  1,  1), 
   vec3( 1,  1, -1), vec3( 1, -1, -1), vec3(-1, -1, -1), vec3(-1,  1, -1),
   vec3( 1,  1,  0), vec3( 1, -1,  0), vec3(-1, -1,  0), vec3(-1,  1,  0),
   vec3( 1,  0,  1), vec3(-1,  0,  1), vec3( 1,  0, -1), vec3(-1,  0, -1),
   vec3( 0,  1,  1), vec3( 0, -1,  1), vec3( 0, -1, -1), vec3( 0,  1, -1)
);
int numSampleDir = 20;   

void main()
{
	vec3 objColor = texture(i_textureUnit, fragmentUV).rgb;
		
	vec3 N = CalcBumpedNormal(); 
	//vec3 N = normalize(v_worldNormal);
	vec3 V = normalize(v_view);
	
	vec3 ambLit = f_matAmbient * objColor * ambLight.v_color;
	vec3 directLit = DirectionalLightsIllumination(N, objColor);
	vec3 pointLit = PointLightsIllumination(V, N, objColor);
	vec3 spotLit = SpotLightsIllumination(V, N, objColor);
	
	vec3 finalColor = clamp(ambLit + directLit + pointLit + spotLit, 0.0f, 1.0f);
	FragColor = vec4(finalColor, f_alpha);
	vec3 brightColor = clamp(finalColor - vec3(0.3, 0.25, 0.25), 0.0f, 1.0f);
	BrightColor = vec4(brightColor, f_alpha);
}

vec3 CalcBumpedNormal()
{
    vec3 Normal = normalize(v_worldNormal);
    vec3 Tangent = normalize(v_tangent);
    Tangent = normalize(Tangent - dot(Tangent, Normal) * Normal);
    vec3 Bitangent = cross(Tangent, Normal);
    vec3 BumpMapNormal = texture(i_normalUnit, fragmentUV).xyz;
    BumpMapNormal = 2.0 * BumpMapNormal - vec3(1.0, 1.0, 1.0);
    vec3 NewNormal;
    mat3 TBN = mat3(Tangent, Bitangent, Normal);
    NewNormal = TBN * BumpMapNormal;
    NewNormal = normalize(NewNormal);
    return NewNormal;
}


vec3 DirectionalLightsIllumination(vec3 N, vec3 objColor)
{
	float cosNL = clamp(dot(N,dirLight.v_direction), 0.0f, 1.0f);

	vec3 finalColor = f_matDiffuse * objColor * dirLight.v_color * cosNL;
	return finalColor;
}

vec3 PointLightsIllumination(vec3 V, vec3 N, vec3 objColor)
{
	vec3 finalColor = vec3(0.0f, 0.0f, 0.0f);
	
	for (int i = 0 ; i < numPointLights ; i++) {
		float shadowFactor = 1.0;
		if(i == 0) {
			
			//omnidirectional shadowMapping
			float shadow = 0.0;
			float eps = 0.15;
			float diskRadius = 0.05;

			vec3 direction = v_position - pointLights[0].v_position;
			float depth_p = length(direction);
			float depth_c;
			for(int i = 0; i < numSampleDir; i++)
			{
				depth_c = texture(i_shadowMapcubeUnit, direction + cubeShadowSamplingDirection[i] * diskRadius).r;
				depth_c = depth_c * v_nearFar.y;
				if(depth_c+eps < depth_p)
					shadow += 1.0;
			}
			shadow /= float(numSampleDir);
			shadowFactor = 1.0 - shadow; 
			
			//withoud PCF
			//vec3 direction = v_position - pointLights[0].v_position;
			//float depth_p = length(direction);
			//float depth_c = texture(i_shadowMapcubeUnit, direction).r;
			//depth_c = depth_c * v_nearFar.y;
			//float eps = 0.15;
			//shadowFactor = (depth_c+eps < depth_p) ? 0.0 : 1.0;
		}

		vec3 L = normalize(pointLights[i].v_position - v_position);
		vec3 I = calcPointLightIntensity(i);

		vec3 R = reflect(L,N)*(-1);
		//vec3 H = normalize(V + L);

		float cosNL = clamp(dot(N,L), 0.0f, 1.0f);
		
		float specularFac = cosNL * pow(clamp(dot(R,V), 0.0f, 1.0f), f_matShininess);
		//float specularFac = cosNL * pow(clamp(dot(N,H), 0.0f, 1.0f), f_matShininess);
		
		vec3 diffPart = f_matDiffuse * objColor * I * cosNL;
		vec3 specPart = f_matSpecular * I * specularFac;

		finalColor += shadowFactor * (diffPart + specPart);
	}
	
	return clamp(finalColor, 0.0f, 1.0f);
}

vec3 SpotLightsIllumination(vec3 V, vec3 N, vec3 objColor)
{
	vec3 finalColor = vec3(0.0f, 0.0f, 0.0f);
	
	for (int i = 0 ; i < numSpotLights ; i++) {

		vec3 L = normalize(spotLights[i].v_position - v_position);
		float spotFactor = dot(L, spotLights[i].v_direction);

		if(spotFactor > spotLights[i].f_cutoff) {
			float shadowFactor = 1.0;
			if(i == 0) {
				//shadowMapping
				//shadowFactor = textureProj(i_shadowMaptextureUnit, v_smPosition);
			}

			vec3 I = calcSpotLightIntensity(i);

			vec3 R = reflect(L,N)*(-1);
			//vec3 H = normalize(V + L);

			float cosNL = clamp(dot(N,L), 0.0f, 1.0f);
		
			float specularFac = cosNL * pow(clamp(dot(R,V), 0.0f, 1.0f), f_matShininess);
			//float specularFac = cosNL * pow(clamp(dot(N,H), 0.0f, 1.0f), f_matShininess);
		
			vec3 diffPart = f_matDiffuse * objColor * I * cosNL;
			vec3 specPart = f_matSpecular * I * specularFac;

			vec3 lightColor = diffPart + specPart;

			//mapped from [cutoff, 1] to [0,1]
			float radialIntensity = (1.0 - (1.0 - spotFactor) * 1.0/(1.0 - spotLights[i].f_cutoff));

			finalColor += shadowFactor * lightColor * radialIntensity;
		}		
	}
	
	return clamp(finalColor, 0.0f, 1.0f);
}


vec3 calcPointLightIntensity(int lightIndex)
{
	float dist = distance(pointLights[lightIndex].v_position, v_position);
	float attenuation = pointLights[lightIndex].v_attenuation.x + 
						pointLights[lightIndex].v_attenuation.y * dist + 
						pointLights[lightIndex].v_attenuation.z * dist * dist;
	
	vec3 intensity = pointLights[lightIndex].v_color / attenuation;
	return intensity;		
}

vec3 calcSpotLightIntensity(int lightIndex)
{
	float dist = distance(spotLights[lightIndex].v_position, v_position);
	float attenuation = spotLights[lightIndex].v_attenuation.x + 
						spotLights[lightIndex].v_attenuation.y * dist + 
						spotLights[lightIndex].v_attenuation.z * dist * dist;
	
	vec3 intensity = spotLights[lightIndex].v_color / attenuation;
	return intensity;		
}